package search;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataOutputStream;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapred.FileInputFormat;
import org.apache.hadoop.mapred.FileOutputFormat;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Reporter;

import utils.RPNUtils;

@SuppressWarnings("deprecation")
public class Searcher {

	private static String[] rpnQuery;
	private static int wordsCount;
	private static JobClient client;
	private static JobConf conf;

	private static FileSystem hdfs1;
	static {
		try {
			hdfs1 = FileSystem.get(new Configuration());
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public static class SearchMapper extends MapReduceBase implements
			Mapper<LongWritable, Text, Text, Text> {

		public void map(LongWritable key, Text val,
				OutputCollector<Text, Text> output, Reporter reporter)
				throws IOException {
			System.out.println("mapper");
			String word = val.toString().trim();

			Path index = new Path("Index");
			FileStatus[] status = hdfs1.listStatus(index);
			boolean found = false;
			for (int i = 0; i < status.length; i++) {
				if (status[i].isDir())
					continue;
				BufferedReader br = new BufferedReader(new InputStreamReader(
						hdfs1.open(status[i].getPath())));
				String line;
				line = br.readLine();
				System.out.println(word);
				while (line != null) {
					if (line.startsWith(word)) {
						output.collect(new Text(word), new Text(line));
						found = true;
						br.close();
						break;
					}
					line = br.readLine();
				}
				if (found){
					br.close();
					break;
				}
			}

			if (!found) {
				output.collect(new Text(word), new Text("null"));
			}
			//hdfs1.close();
		}
	}

	public static class SearchReducer extends MapReduceBase implements
			Reducer<Text, Text, Text, Text> {

		public HashMap<String, Set<String>> acc = new HashMap<String, Set<String>>();
		public Stack<Set<String>> result = new Stack<Set<String>>();
		public String[] query;
		public int wordCount;
		public int counter = 0;

		public void reduce(Text key, Iterator<Text> values,
				OutputCollector<Text, Text> output, Reporter reporter)
				throws IOException {

			Set<String> docs = new HashSet<String>();
			while (values.hasNext()) {
				String[] results = values.next().toString().split("(\\s|\\t)");
				for (String result : results) {
					if (result.equals("") || result.equals(key.toString()))
						continue;
					else
						docs.add(result);
				}
			}
			acc.put(key.toString(), docs);
			++counter;
			if (counter == wordCount) {
				Set<String> res = RPNUtils.compute(acc, query);
				output.collect(new Text("result"), new Text(res.toString()));
			}
			//output.collect(new Text("result"+" "+Searcher.wordsCount+" "+counter+" "+Arrays.toString(rpnQuery)), new Text(acc.toString()));
		}
		
		@Override
		public void configure(JobConf job) {
			wordCount = job.getInt("wordcount", 0);
			query = job.getStrings("query");
		}
	}


	public static void run(String query) throws IOException {

		boolean clean = cleanUp();
		if (clean) {
			processQuery(query);
			configureJob();
			try {
				JobClient.runJob(conf);
			} catch (Exception e) {
				e.printStackTrace();
			}

		}

	}

	private static boolean cleanUp() {
		boolean isQueryDeleted;
		boolean isResultDeleted;

		Path queryPath = new Path("query");
		Path result = new Path("result");
		try {
			FileSystem hdfs = FileSystem.get(new Configuration());
			if (!hdfs.exists(queryPath)) {
				isQueryDeleted = true;
			} else {
				isQueryDeleted = hdfs.delete(queryPath, true);
			}
			if (!hdfs.exists(result)) {
				isResultDeleted = true;
			} else {
				isResultDeleted = hdfs.delete(result, true);
			}
			hdfs.close();
		} catch (IOException e) {
			return false;
		}

		return isQueryDeleted && isResultDeleted;
	}

	private static void configureJob() {
		client = new JobClient();
		conf = new JobConf(Searcher.class);

		conf.setJobName("search");
		conf.setOutputKeyClass(Text.class);
		conf.setOutputValueClass(Text.class);
		conf.setInt("wordcount", wordsCount);
		conf.setStrings("query", rpnQuery);
		conf.setNumReduceTasks(1);

		FileInputFormat.addInputPath(conf, new Path("query"));
		FileOutputFormat.setOutputPath(conf, new Path("result"));

		conf.setMapperClass(SearchMapper.class);
		conf.setReducerClass(SearchReducer.class);

		client.setConf(conf);
	}

	private static void processQuery(String query) {
		String rpnString = RPNUtils.infixToRPN(query);
		rpnQuery = rpnString.split(" ");
		List<String> words = RPNUtils.parseQuery(rpnString);
		wordsCount = words.size();
		FileSystem hdfs = null;
		try {
			hdfs = FileSystem.get(new Configuration());
		} catch (IOException e) {
			e.printStackTrace();
		}
		for (int i = 0; i < words.size(); ++i) {
			Path path = new Path("query/" + words.get(i));
			FSDataOutputStream outputStream = null;
			try {
				outputStream = hdfs.create(path);
				outputStream.write(words.get(i).getBytes(), 0, words.get(i)
						.getBytes().length);
				outputStream.close();
			} catch (IOException e1) {
				e1.printStackTrace();
			}
		}
		try {
			hdfs.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

}